/*
 * ssidsniff - a tool for wireless network reconnaissance
 * Copyright (C) 2002-2007 Kostas Evangelinos <kos@bastard.net>
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*
 * $Id: 80211.c,v 1.31 2007-06-28 14:28:01 kos Exp $
 * 
 */

#undef DEBUG_PARSER

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <limits.h>
#include <string.h>

#ifdef HAVE_PCAP_H
#include "pcap.h"
#endif

#ifdef HAVE_PCAP_BPF_H
#include "pcap-bpf.h"
#endif

#include "ssidsniff.h"
#include "llog.h"
#include "util.h"
#include "80211.h"
#include "support.h"

static int
f80211_parse_tags(unsigned char *p, unsigned int l, f80211_t *res) {
  unsigned int le;
  unsigned int tlen, ttype;

  for(le=0; le <= l; ) {
      
    if(le+1 >= l)
      break;

    ttype=(unsigned char)EXTRACT_LE_8BITS(p+le);
    tlen=(unsigned char)EXTRACT_LE_8BITS(p+le+1);

    if(le+1+tlen > l) {
      llog(6, "parse: problem decoding tag, ignoring (%d+1+%d)>%d\n", le, tlen, l);
      break;
    }

    switch(ttype) {
    case TAG_SSID:      
      if(*res->ssid) {
	llog(2, "  parse: Found duplicate SSID tag\n");
	break;
      }
      memset(res->ssid, '\0', DEF_LINELEN);
      memcpy(res->ssid, &p[le+2], tlen);
#ifdef DEBUG_PARSER
      llog(4, "   tag SSID: %s\n", res->ssid);
#endif
      break;
    case TAG_SUPP_RATES:
#ifdef DEBUG_PARSER
      llog(4, "   tag RATES: \n");
#endif
      break;
    case TAG_DS_PARAMETER:
#ifdef DEBUG_PARSER
      llog(4, "   tag DSPARAMS: ");
#endif
      res->channel = EXTRACT_LE_8BITS(p+le+2);
#ifdef DEBUG_PARSER
      llog(4, "channel %d\n", res->channel);
#endif
      break;
    case TAG_IBSS_PARAMETER:
#ifdef DEBUG_PARSER
      llog(4, "   tag IBSSPARAMS:\n");
#endif
      break;
    case TAG_EXTENDED_RATES:
#ifdef DEBUG_PARSER
      llog(4, "   tag EXTENDED_RATES:\n");
#endif
      res->xrates = 1;
      break;
    case TAG_VENDOR_SPECIFIC:
#ifdef DEBUG_PARSER
      llog(4, "   tag VENDOR_SPECIFIC: %x:%x:%x\n", p[le+2], p[le+3], p[le+4]);
#endif
      if(!memcmp(p+le+2, WPA_OUI"\x01", 4))
        res->wpa = 1;
      break;
    case TAG_WPA2:
#ifdef DEBUG_PARSER
      llog(4, "   tag WPA2:\n");
#endif
      res->wpa2 = 1;
      break;
    default:
#ifdef DEBUG_PARSER
      llog(4, "   unhandled tag %u, len %u\n", ttype, tlen);
#endif
      break;
    }
    le += tlen + 2;
  }  
  return 0;
}

int
radiotap_process(unsigned char *p, struct pcap_pkthdr *hdr, f80211_t *res, int *off) {
  struct ieee80211_radiotap_header *rh;
  u_int32_t i;
  int len;
  int present;
  int version;

  rh = (struct ieee80211_radiotap_header *)p;
  len = EXTRACT_LE_16BITS(&rh->it_len);
  version = EXTRACT_LE_8BITS(&rh->it_version);
  present = EXTRACT_LE_32BITS(&rh->it_present);
#ifdef DEBUG_PARSER
  llog(4, "radiotap: version %d, len %d, present %d\n", 
       version, len, present);
#endif

  if(hdr->caplen < len) {
    llog(3, " parse: packet truncated, discarding\n");
    return 3;
  }

  *off = len;
  if((1<<IEEE80211_RADIOTAP_EXT) & present) {
    llog(4, " extended bit set. Bailing\n");
    return 1;
  }

  // XXX: implement alignment cursor
  p += sizeof(struct ieee80211_radiotap_header);
  for(i=0; i<IEEE80211_RADIOTAP_DB_ANTNOISE; i++) {
    if((1<<i) & present) {
#ifdef DEBUG_PARSER
      llog(4, " field %d: present\n", i);
#endif
      switch(i) {
      case IEEE80211_RADIOTAP_TSFT:
	p+=8;
	break;
      case IEEE80211_RADIOTAP_FLAGS:
	p+=1;
	break;
      case IEEE80211_RADIOTAP_RATE:
	p+=1;
	break;
      case IEEE80211_RADIOTAP_CHANNEL:
	p+=4;
	break;
      case IEEE80211_RADIOTAP_FHSS:
	p+=2;
	break;
      case IEEE80211_RADIOTAP_DBM_ANTSIGNAL:
	res->signal = (char)EXTRACT_LE_8BITS(p);
#ifdef DEBUG_PARSER
	llog(4, " signal %d\n", res->signal);
#endif
	break;
      }
    }
  }
  return 0;
}

int
prism2_process(unsigned char *p, struct pcap_pkthdr *hdr, f80211_t *res, int *off) {
  struct prism2_hdr *ph;
  int len;

  ph = (struct prism2_hdr *)p;
  if(hdr->caplen < PRISM2_LEN) { 
    llog(3, "parse: packet truncated, discarding\n");
    return 3;
  }
  len = EXTRACT_LE_16BITS(&ph->msglen);
  *off = len;
  res->signal = EXTRACT_LE_32BITS(&ph->signal.data);
  return 0;
}

int
f80211_parse(unsigned char *p, struct pcap_pkthdr *hdr, int dlt, f80211_t *res) {
  f80211_mac_t *mac;
  frame_control_t *fc;
  beacon_header_t *bh;
  int offset = 0;
  
  if(hdr->caplen != hdr->len) {
    llog(3, "parse: packet truncated, discarding\n");
    return 2;
  }

  /* this happens quite often for LLC packets, not interesting */  
  if(hdr->len < sizeof(f80211_mac_t)) {
#ifdef DEBUG_PARSER
    llog(3, "parse: packet too small to have 802.11 MAC header\n");
#endif
    return 1;
  }

  memset(res, '\0', sizeof(f80211_t));
  res->signal = CHAR_MIN;

  if(dlt == DLT_IEEE802_11_RADIO) {
#ifdef DEBUG_PARSER
    llog(4, "Parsing radiotap header based on DLT\n");
#endif
    if(radiotap_process(p, hdr, res, &offset))
      return 10;
    p += offset;
  } 

  if(dlt == DLT_PRISM_HEADER) {
#ifdef DEBUG_PARSER
    llog(4, "Parsing prism header based on DLT\n");
#endif
    if(prism2_process(p, hdr, res, &offset))
      return 11;
    p += offset;
  }

#ifdef DEBUG_PARSER
  llog(4, "parse: Offset skip due to radio headers is %d\n", offset);
#endif
  mac = (f80211_mac_t *)p;
  memcpy(&res->f_mac, mac, sizeof(f80211_mac_t));
  fc = (frame_control_t *)mac->fc;
  if(fc->version) {
    llog(3, "parse: unhandled version number 0x%x\n", fc->version);
    return 2;
  }

#ifdef DEBUG_PARSER
  llog(4, "MAC 80211\n FC: version %u, type %u, subtype %u\n", 
       fc->version, fc->type, fc->subtype);
#endif

  res->type = fc->type;
  res->subtype = fc->subtype;
  res->wepused = fc->wep;   /* wepused for _this_ packet */

  if(fc->type == DATA_FRAME) {

/*
 *  Data Frame - Address field contents
 *
 *  To Ds  | From DS | Addr 1 | Addr 2 | Addr 3 | Addr 4
 *    0    |  0      |  DA    | SA     | BSSID  | n/a
 *    0    |  1      |  DA    | BSSID  | SA     | n/a
 *    1    |  0      |  BSSID | SA     | DA     | n/a
 *    1    |  1      |  RA    | TA     | DA     | SA
 */

    if(fc->to_ds && !fc->from_ds) {
      strncpy(res->bss, mac2ascii(mac->address1), DEF_LINELEN);
      strncpy(res->src, mac2ascii(mac->address2), DEF_LINELEN);
      strncpy(res->dst, mac2ascii(mac->address3), DEF_LINELEN);
    } else if(fc->from_ds && !fc->to_ds) {
      strncpy(res->dst, mac2ascii(mac->address1), DEF_LINELEN);
      strncpy(res->bss, mac2ascii(mac->address2), DEF_LINELEN);
      strncpy(res->src, mac2ascii(mac->address3), DEF_LINELEN);
    } else if(!fc->from_ds && !fc->to_ds) {
      strncpy(res->dst, mac2ascii(mac->address1), DEF_LINELEN);
      strncpy(res->src, mac2ascii(mac->address2), DEF_LINELEN);
      strncpy(res->bss, mac2ascii(mac->address3), DEF_LINELEN);
    } else {
      /* unhandled */
      return 0;
    }
#ifdef DEBUG_PARSER
    llog(4, " Type: data frame, subtype %d\n", fc->subtype);
    llog(4, " BSS %s, SA %s DA %s\n", res->bss, res->src, res->dst);
#endif
    return 0;
  }

  if(fc->type != MGT_FRAME) {
    llog(4, " Non-management packet type %d subtype %d ignored\n", fc->type, fc->subtype);
    return 0;
  }

  switch(fc->subtype) {
  case MGT_PROBE_REQ:
    strncpy(res->dst, mac2ascii(mac->address1), DEF_LINELEN);
    strncpy(res->src, mac2ascii(mac->address2), DEF_LINELEN);
    strncpy(res->bss, mac2ascii(mac->address3), DEF_LINELEN);
#ifdef DEBUG_PARSER
    llog(4, " Type: Probe request\n");
    llog(4, " BSS %s, SA %s DA %s\n", res->bss, res->src, res->dst);
#endif
    f80211_parse_tags(p + sizeof(f80211_mac_t),
		      hdr->len - (sizeof(f80211_mac_t)+offset),
		      res);
    break;

  case MGT_PROBE_RESP:
  case MGT_BEACON:
    strncpy(res->dst, mac2ascii(mac->address1), DEF_LINELEN);
    strncpy(res->src, mac2ascii(mac->address2), DEF_LINELEN);
    strncpy(res->bss, mac2ascii(mac->address3), DEF_LINELEN);
    if(sizeof(beacon_header_t)+sizeof(f80211_mac_t) > hdr->len) {
#ifdef DEBUG_PARSER
      llog(4, "parse: bad frame size\n");
#endif
      return 3;
    }

    bh = (beacon_header_t *)(p+sizeof(f80211_mac_t));
#ifdef DEBUG_PARSER
    if(fc->subtype == MGT_BEACON)
      llog(4, " Type: Beacon\n");
    else
      llog(4, " Type: Probe response\n");
    llog(4, " BSS %s, SA %s DA %s\n", res->bss, res->src, res->dst);
    llog(4, "  timestamp %s, interval %s, cap %s\n", 
	 hex2ascii(bh->timestamp, 8),
	 hex2ascii(bh->interval, 2),
	 hex2ascii(bh->capability, 2));    
#endif

    /* fixed parameters */
    res->ap  = bh->capability[0] & IEEE80211_CAPINFO_ESS;
    res->wepcapable = bh->capability[0] & IEEE80211_CAPINFO_PRIVACY;
    res->interval = bh->interval[0] + (bh->interval[1]<<8);
    
    /* tagged parameters */
    f80211_parse_tags(p + (sizeof(f80211_mac_t)+sizeof(beacon_header_t)), 
		      hdr->len - (sizeof(f80211_mac_t)+sizeof(beacon_header_t)+offset),
		      res);
    break;
  default:
    llog(3, "parse: unknown mgmt frame subtype %d\n", *p>>4);
    return 4;
    break;
  }

  return 0;
}

int
f80211_isbeacon(f80211_t *frame) {
  return(frame->type == 0 && frame->subtype == 8);
}

/* data packet (wep or cleartext) */
int
f80211_isdata(f80211_t *frame) {
  return(frame->type == DATA_FRAME &&                            
	    (frame->subtype == 0 || 
	     frame->subtype == 1 || 
	     frame->subtype == 2 || 
	     frame->subtype == 3));
}

int
f80211_isprobereq(f80211_t *frame) {
  return(frame->type == MGT_FRAME && frame->subtype==MGT_PROBE_REQ);
}

int
f80211_isproberesp(f80211_t *frame) {
  return(frame->type == MGT_FRAME && frame->subtype==MGT_PROBE_RESP);
}

f80211_frame_t
f80211_type(f80211_t *frame) {
  if(frame->type == MGT_FRAME && frame->subtype == MGT_BEACON)
    return f80211_BEACON;
  else if(frame->type == MGT_FRAME && frame->subtype==MGT_PROBE_REQ)
    return f80211_PROBEREQ;
  else if(frame->type == MGT_FRAME && frame->subtype==MGT_PROBE_RESP)
    return f80211_PROBERESP;
  else if(frame->type == DATA_FRAME &&                            
	    (frame->subtype == 0 || 
	     frame->subtype == 1 || 
	     frame->subtype == 2 || 
	     frame->subtype == 3))
    return f80211_DATA;
  else
    return f80211_UNKNOWN;
}

int
f80211_checkdlt(int dlt) {
  if(dlt != DLT_IEEE802_11 &&
     dlt != DLT_IEEE802_11_RADIO &&
     dlt != DLT_PRISM_HEADER)
    return 1;
  return 0;
}
