/*
 * ssidsniff - a tool for wireless network reconnaissance
 * Copyright (C) 2002-2007 Kostas Evangelinos <kos@bastard.net>
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*
 * $Id: cursesui.c,v 1.69 2007-06-28 14:28:01 kos Exp $
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <limits.h>
#include <sys/param.h>
#include <ncurses/ncurses.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "cursesui.h"
#include "util.h"
#include "mach.h"
#include "ouidb.h"
#include "hopper.h"

#ifdef HAVE_PCAP_H
#include "pcap.h"
#endif

#ifdef HAVE_NET_BPF_H
#include "net/bpf.h"
#endif

extern char pcap_version[];

#define UI_STATUS_POS 0
#define UI_QUERY_POS 2
#define UI_HEADER_POS 3
#define UI_MAIN_POS 4

static char rotseq[] = "/-\\|";
static int disp_mode;
static int sort_mode;
static int selected_network;
static char line_info[DEF_LINELEN];
static time_t line_info_update;
static char line_query[DEF_LINELEN];
static char query_resc;
static char query_ress[DEF_LINELEN];
static ui_query_t query_type;
static int disp_offset;
static time_t full_update;

/* 
 * XXX: Order is important until this is done better.
 */

uifield_t uifield[] = {
  {"SSID Name", 22, 0},
  {"BSSID", 19, 0},
  {"Flg", 6, 0},
  {"Cn", 3, 0},
  {"Pkts", 7, 0},
  {"Data", 7, 0},
  {"PPS", 5, 0},
  {"Sig", 4, 0},
  {"", 0, 0, 0},
};

char *
status_flag_create(wnet_t *sample) {
  static char status[DEF_LINELEN];
  int i=0;
  
  status[i++] = (sample->ap?'A':'H');
	if(sample->wepused)
		status[i++] = 'E';
	else if(sample->wpa || sample->wpa2)
		status[i++] = 'w';
	else if (sample->wepcapable)
		status[i++] = 'e';

  if(sample->xrates)
    status[i++] = 'x';
  if(sample->capture)
    status[i++] = 'W';
  status[i++] = '\0';
  return status;
}

char *
global_status(ssidsniff_t *s) {
  static char status[DEF_LINELEN];
  
  status[0] = '\0';
  if(s->pd_dump) 
    strncat(status, "Cap", DEF_LINELEN);

  if(s->capture == CAPTURE_DATA)
    strncat(status, "Data", DEF_LINELEN);
  else if(s->capture == CAPTURE_ALL)
    strncat(status, "All", DEF_LINELEN);

  if(s->geiger)
    strncat(status, "Snd", DEF_LINELEN);
  
  if(s->run_hopper)
    strncat(status, "Hop", DEF_LINELEN);

  return status;
}

int
field_get_value(int field_id, wnet_t *sample, char *res, int resl) {
  
  *res = '\0';
  switch(field_id) {
  case 0:
    snprintf(res, resl, "%-26s", string_printable(sample->ssid));
    break;
  case 1:
    snprintf(res, resl, "%-20s", sample->bss);
    break;
  case 2:
    snprintf(res, resl, "%-5s", status_flag_create(sample));
    break;
  case 3:
    if(sample->channel)
      snprintf(res, resl, "%-4d", sample->channel);
    break;
  case 4:
    if(sample->count.total)
      snprintf(res, resl, "%-7s", geek_print(sample->count.total));
    break;
  case 5:
    if(sample->count.data)
      snprintf(res, resl, "%-6s", geek_print(sample->count.data));
    break;
  case 6:
    if(sample->pps)
      snprintf(res, resl, "%-4s", geek_print(sample->pps));
    break;
  case 7:
    if(sample->signal != CHAR_MIN)
      snprintf(res, resl, "%-4d", sample->signal);
    break;
  }
  return 0;
}

static int
sortfunc(wnet_t *a, wnet_t *b) {
  switch(sort_mode) {
  case UI_SORT_RECV:
    return (a->serial > b->serial);
  case UI_SORT_SSID:
    return strcmp(a->ssid, b->ssid);
    break;
  case UI_SORT_CHANNEL:
    return (a->channel < b->channel);
    break;
  case UI_SORT_PACKETS:
    return (a->count.total < b->count.total);
    break;
  case UI_SORT_DATA:
    return (a->count.data < b->count.data);
    break;
  case UI_SORT_PPS:
    return (a->pps < b->pps);
    break;
  case UI_SORT_BSSID:
    return strcmp(a->bss, b->bss);
    break;
  case UI_SORT_SIGNAL:
    return (a->signal < b->signal);
    break;
  case UI_SORT_ACTIVE:
    return (a->last_activity.tv_sec < b->last_activity.tv_sec);
    break;
  }
  return 0;
}

static int
header_show(ssidsniff_t *s) {
  static int spindex = 0;
  static int total;
  float freq;

  freq = mach_freq_get(s->device);
  move(UI_STATUS_POS, 0);
  printw("%s %c up %s  %d networks  %s  %s", 
	 PACKAGE, rotseq[(spindex)%4],
	 uptime(s->start), s->num_networks, mach_battery_status(),
	 global_status(s));
  move(UI_STATUS_POS+1, 0);
  printw("%s: ", s->device);
  if(freq > 0)
    printw("%.3fGHz, ", mach_freq_get(s->device)/(double)(1000000000));
  printw("%s pkts: ", geek_print(s->globalcount.total));
  printw("%s beacons, ", geek_print(s->globalcount.beacon));
  printw("%s data, ", geek_print(s->globalcount.data));
  printw("%s other", geek_print(s->globalcount.martian));
  move(UI_STATUS_POS+2, 0);

  if(disp_mode == UI_MODE_NORMAL)
    printw("%s", line_info);

  if(total != s->globalcount.total)
    spindex++;
  total = s->globalcount.total;

  return 0;
}

static int
list_show(ssidsniff_t *s) {
  wnet_t *network;
  int i;
  int field_id;
  int column;
  char header[DEF_LINELEN];
  char format[DEF_LINELEN];
  char value[DEF_LINELEN];
  int network_index;
  
  qsort(s->networks, s->num_networks, sizeof(wnet_t), 
	(int(*)(const void *, const void *))sortfunc);
  
  // blank header and figure out the fields we can show
  header[0] = '\0';
  column=1;
  for(field_id=0; uifield[field_id].size; field_id++) {
    if(uifield[field_id].size + column > COLS) {
      uifield[field_id].cursize = 0;
      break;
    }
    column += uifield[field_id].size;
    uifield[field_id].cursize = uifield[field_id].size;
  }
  
  // print header
  move(UI_HEADER_POS, 0);
  attron(A_REVERSE);
  for(i=0; i<COLS; i++)
    printw(" ");
  move(UI_HEADER_POS, 1);
  for(field_id=0; uifield[field_id].cursize; field_id++) {
    snprintf(format, DEF_LINELEN, "%%-%ds", uifield[field_id].cursize);
    printw(format, uifield[field_id].name);
  }
  attroff(A_REVERSE);  

  // print networks 
  for(i=0; i < LINES-UI_MAIN_POS; i++) {
    network_index = disp_offset + i;
    if(network_index >= s->num_networks)
      break;
    
    network = &s->networks[network_index];

    move(UI_MAIN_POS+i, 0);
    if(network_index == selected_network)
      printw(">");
    else
      printw(" ");

    column = 1;
    for(field_id=0; uifield[field_id].cursize; field_id++) {
      move(UI_MAIN_POS+i, column);
      field_get_value(field_id, network, value, DEF_LINELEN);
      printw(value);
      column += uifield[field_id].size;
    }
  }
  move(UI_STATUS_POS, COLS-1);
  return 0;
}

static int
info_show(ssidsniff_t *s) {
  wnet_t *n;
  wnet_client_t *c;
  int i;

  n = &s->networks[selected_network];
  
  move(UI_HEADER_POS, 0);
  attron(A_REVERSE);
  for(i=0; i<COLS; i++)
    printw(" ");
  move(UI_HEADER_POS, 0);
  printw("Network Information for %s", n->bss);
  attroff(A_REVERSE);
  
  move(UI_MAIN_POS, 0);
  printw("SSID: %s\n", n->ssid);
  printw("Packet count: %u (beacon %u, data %u, unk %u)\n", 
   n->count.total, n->count.beacon, n->count.data, n->count.martian);
  printw("Network members: \n");
  for(c=n->clients; c; c=c->next)
    printw(" [%s] %s (%s) %s packets\n", c->ap?"A":"H", c->src, c->manufacturer, geek_print(c->count));
  printw("Channel: %d\n", n->channel);
  printw("Access point: %s\n", n->ap?"Yes":"No");
  printw("Packets per second: %u\n", n->pps);
  printw("Encryption capabilities: %s %s %s\n", n->wepcapable?"WEP":"", n->wpa?"WPA":"", n->wpa2?"WPA2":"");
  printw("Encryption being used: %s\n", n->wepused?"Yes":"No");
  printw("Last seen activity: %s\n", human_time_delta(&n->last_activity));
  printw("Serial number: %d\n", n->serial);
  move(UI_STATUS_POS, COLS-1);
  return 0;
}

static int
help_show(ssidsniff_t *s) {
  int i;

  move(UI_HEADER_POS, 0);
  attron(A_REVERSE);
  for(i=0; i<COLS; i++)
    printw(" ");
  move(UI_HEADER_POS, 0);
  printw("Help Page | %s %s, libpcap %s", PACKAGE, VERSION, pcap_version);
  attroff(A_REVERSE);  
  
  move(UI_MAIN_POS, 0);
  //----------------------------------------------------------------------------
  printw("\
Space  Update screen\n\
i      Information for currently selected network\n\
j,k    Move network selector\n\
d      Display strings in data packets\n\
o      Select sort order: \n\
           as (r)eceived, (s)sid, (c)hannel, (p)acket count, (d)ata count, \n\
           packets per seco(n)d, (b)ssid, (S)ignal, latest (a)ctivity\n\
w      Start/stop saving data packets from selected network\n\
W      Start/stop saving all data packets received\n\
A      Start/stop saving all packets including MAC layer\n\
s      Geiger mode on/off\n\
c      Set wireless channel\n\
r      Record all wireless networks seen to text file\n\
H      Start/stop channel hopping\n\
f      Focus on a network; stop channel hopper, reset channel\n\
^L     Redraw screen\n\
q      Quit\n\
h      Help\n\
\n");
  printw("Packet saving is %s, file is %s\n", s->pd_dump?"On":"Off", s->dumpfile);
  //  printw("Hardware address db date: %s\n", ouidb_date());
  move(UI_STATUS_POS, COLS-1);
  return 0;
}

static int
strings_show(wnet_t *net) {
  int i;
  int len;

  move(UI_HEADER_POS, 0);
  attron(A_REVERSE);
  for(i=0; i<COLS; i++)
    printw(" ");
  move(UI_HEADER_POS, 0);
  printw("Data strings for %s (%s)", net->ssid, net->bss);
  attroff(A_REVERSE);  
  move(UI_MAIN_POS, 0);

  len = sizeof(net->strings.strings);
  for(i=0; i<len; i++)
    addch(net->strings.strings[i]);
  move(UI_STATUS_POS, COLS-1);
  return 0;
}

int
ui_init(void) {  
  disp_mode = UI_MODE_NORMAL;
  sort_mode = UI_SORT_RECV;
  initscr();
  cbreak();
  noecho();
  keypad(stdscr, TRUE);
  line_info[0] = '\0';
  disp_offset = 0;
  full_update = 0;
  return 0;
}

int
ui_info(char *fmt, ...) {
  va_list vl;

  va_start(vl, fmt);
  vsnprintf(line_info, DEF_LINELEN, fmt, vl);
  va_end(vl);
  return 0;
}

int
ui_query(char *s, ui_query_t type) {
  query_type = type;
  strncpy(line_query, s, DEF_LINELEN);
  strncat(line_query, "? ", DEF_LINELEN);
  return 0;
}

int
ui_mode(ui_mode_t umode) {
  disp_mode = umode;
  return 0;
}

int
ui_finish(void) {
  endwin();
  return 0;
}

int
ui_clear(void) {
  erase();
  return 0;
}

int
ui_sort(ssidsniff_t *s, ui_sort_t mode) {
  sort_mode = mode;
  return 0;
}

int
ui_page(int num_networks, int where) {
  int len = LINES-UI_MAIN_POS;

  switch(where) {
  case UI_DIR_UP:
    if(disp_offset - len >= 0) {
      disp_offset-=len;
      selected_network = disp_offset + (len-1);
    } else if(disp_offset == 0 && selected_network != 0) {
      selected_network = 0; 
    }
    break;
  case UI_DIR_DOWN:
    if(num_networks - disp_offset >= len) {      
      disp_offset+=len;
      selected_network = disp_offset;
    } else if(num_networks - disp_offset < len && selected_network != num_networks-1) {
      selected_network = num_networks-1;
    }
    break;
  }
  return 0;
}

int
ui_select(ssidsniff_t *s, int dir) {
  int lim;
  int len = LINES-UI_MAIN_POS;
  int pos;

  pos = selected_network - disp_offset;
  if(selected_network < disp_offset ||
     selected_network > disp_offset + len)
    selected_network = disp_offset;
  
  lim = MIN(LINES-UI_MAIN_POS-1, s->num_networks-disp_offset-1);

  if(dir == UI_DIR_UP && pos > 0) {
    move(UI_MAIN_POS + pos, 0);
    printw(" ");
    selected_network--; pos--;
    move(UI_MAIN_POS + pos, 0);
    printw(">");
  } else if(dir == UI_DIR_UP && pos == 0 && disp_offset > 0) {
    ui_page(s->num_networks, 0);
    ui_redraw(s);
  } else if(dir == UI_DIR_DOWN && pos < lim) {
    move(UI_MAIN_POS + pos, 0);
    printw(" ");
    selected_network++; pos++;
    move(UI_MAIN_POS + pos, 0);
    printw(">");
  } else if(dir == UI_DIR_DOWN && pos == lim) {
    ui_page(s->num_networks, 1);
    ui_redraw(s);
  } 
  return 0;
}

int
ui_process_events(ssidsniff_t *s) {
  int need_redraw = 0;
  char temp[1024];
  int channel;
  
  switch(disp_mode) {

  case UI_MODE_NORMAL:
    switch(getch()) {
    case KEY_NPAGE:
      ui_page(s->num_networks, UI_DIR_DOWN);
      need_redraw = 1;
      break;
    case KEY_PPAGE:
      ui_page(s->num_networks, UI_DIR_UP);
      need_redraw = 1;
      break;
    case 014: /* ^L */
      clear();
      need_redraw = 1;
      break;
    case 'q': /* quit */
      s->running = 0;
      break;
    case 'h':
    case '?': /* help */
      disp_mode = UI_MODE_HELP;
      need_redraw = 1;
      break;
    case 'i': /* network info */
    case 10:
      disp_mode = UI_MODE_INFO;
      need_redraw = 1;
      break;
    case 'c': /* channel */
      disp_mode = UI_MODE_FREQ;
      need_redraw = 1;
      break;
    case ' ': /* refresh */      
      need_redraw = 1;
      break;
    case 'j': /* down */
    case KEY_DOWN:
      ui_select(s, UI_DIR_DOWN);
      break;
    case 'k': /* up */
    case KEY_UP:
      ui_select(s, UI_DIR_UP);
      break;
    case 'o': /* sort order */
      disp_mode = UI_MODE_SORT;
      need_redraw = 1;
      break;
    case 'w': /* capture data on this network */
      s->networks[selected_network].capture = !s->networks[selected_network].capture;
      break;
    case 'W': /* capture all data */
      if(s->capture == CAPTURE_NONE)
	s->capture = CAPTURE_DATA;
      else
	s->capture = CAPTURE_NONE;
      break;
    case 'A': /* capture all packets */
      if(s->capture == CAPTURE_NONE)
	s->capture = CAPTURE_ALL;
      else
	s->capture = CAPTURE_NONE;
      break;
    case 'd': /* strings */
      disp_mode = UI_MODE_STRINGS;
      need_redraw = 1;
      break;
    case 's': /* Sound */
      s->geiger = !s->geiger;
      break;
    case 'r': /* Record */
      wnet_record(s);
      break;
    case 'H': /* Hopper */
      s->run_hopper = !s->run_hopper;
      if(s->run_hopper)
	hopper_msg(s->hopper, "start");
      else
	hopper_msg(s->hopper, "stop");
      break;
    case 'f':
      s->run_hopper = 0;
      hopper_msg(s->hopper, "stop");
      channel = s->networks[selected_network].channel;
      if(channel) {
	mach_channel_set(s->device, channel);
	ui_info("Focusing on channel %d", channel);
	need_redraw = 1;
      }
      break;
/*     case 'S': /\* speed *\/ */
/*       disp_mode = UI_MODE_QUERYS; */
/*       ui_query("What is the speed", QUERY_STRING); */
/*       need_redraw = 1; */
/*       break; */
    }
    break;
  case UI_MODE_QUERYC:
    move(UI_QUERY_POS, 0);
    printw(line_query);
    query_resc = getch();
    disp_mode = UI_MODE_NORMAL;
    break;
  case UI_MODE_QUERYS:
    move(UI_QUERY_POS, 0);
    printw(line_query);
    echo();
    getnstr(query_ress, DEF_LINELEN);
    noecho();
    disp_mode = UI_MODE_NORMAL;
    break;
  case UI_MODE_HELP:
  case UI_MODE_INFO:
  case UI_MODE_STRINGS:
    getch();
    disp_mode = UI_MODE_NORMAL;
    need_redraw = 1;
    break;
  case UI_MODE_SORT:
    switch(getch()) {
    case 'r': /* recv */
      ui_sort(s, UI_SORT_RECV);
      break;
    case 's': /* ssid */
      ui_sort(s, UI_SORT_SSID);
      break;
    case 'c': /* channel */
      ui_sort(s, UI_SORT_CHANNEL);
      break;
    case 'p': /* packets */
      ui_sort(s, UI_SORT_PACKETS);
      break;
    case 'd': /* data packets */
      ui_sort(s, UI_SORT_DATA);
      break;
    case 'n': /* PPS */
      ui_sort(s, UI_SORT_PPS);
      break;
    case 'b': /* BSSID */
      ui_sort(s, UI_SORT_BSSID);
      break;
    case 'S': /* signal */
      ui_sort(s, UI_SORT_SIGNAL);
      break;
    case 'a': /* activity */
      ui_sort(s, UI_SORT_ACTIVE);
      break;
    default:
      break;
    }
    disp_mode = UI_MODE_NORMAL;
    need_redraw = 1;
    break;
  case UI_MODE_FREQ:
    echo();
    getnstr(temp, 2);
    noecho();
    if(atoi(temp))
      mach_channel_set(s->device, atoi(temp));
    disp_mode = UI_MODE_NORMAL;
    need_redraw = 1;
    break;
  }

  if(need_redraw)
    ui_redraw(s);
  else
    ui_refresh(s);
  
  return 0;
}

int
ui_refresh(ssidsniff_t *s) {

  refresh();
  return 0;
}

int
ui_redraw(ssidsniff_t *s) {
  time_t now;
  static int freeze_freq;

  now = time(NULL);

  if(disp_mode != UI_MODE_FREQ) {
    freeze_freq = 0;
    
    if(now - full_update > 20) {
      clear();
      full_update = now;
    }
  }

  switch(disp_mode) {
  case UI_MODE_NORMAL:
    ui_clear();
    header_show(s);
    list_show(s);
    break;
  case UI_MODE_INFO:
    ui_clear();  
    header_show(s);  
    info_show(s);
    break;
  case UI_MODE_HELP:
    ui_clear();
    header_show(s);
    help_show(s);
    break;
  case UI_MODE_SORT:
    ui_clear();
    header_show(s);
    list_show(s);
    move(UI_QUERY_POS, 0);
    printw("Sort order? ");
    break;
  case UI_MODE_STRINGS:
    ui_clear();
    header_show(s);
    strings_show(&s->networks[selected_network]);
    break;
  case UI_MODE_FREQ:
    if(!freeze_freq) {
      ui_clear();
      header_show(s);
      list_show(s);
      move(UI_QUERY_POS, 0);
      printw("Channel? ");
      freeze_freq = 1;
    }
    break;
  case UI_MODE_QUERYC:
  case UI_MODE_QUERYS:
    break;
  }

  /* periodically remove old status messages */
  if(now - line_info_update > 3) {
    ui_info("");
    line_info_update = now;
  }

  refresh();
  return 0;
}
