/*
 * ssidsniff - a tool for wireless network reconnaissance
 * Copyright (C) 2002-2007 Kostas Evangelinos <kos@bastard.net>
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*
 * $Id: hopper.c,v 1.11 2007-06-27 03:57:49 kos Exp $
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <errno.h>

#include "util.h"
#include "xmalloc.h"
#include "mach.h"
#include "hopper.h"
#include "llog.h"

static void
reset_bonus(hopper_t *h) {
  int ch;

  for(ch=0; ch<h->nchannels; ch++)
    h->channels[ch].bonus = 0;
  llog(3, "hopper: resetting timeslice bonuses\n");  
}

static void
parse_command(hopper_t *h, char *line) {

  if(!strncmp(line, "stop", 4)) {
    reset_bonus(h);
    h->running = 0;
  }

  if(!strncmp(line, "start", 5))
    h->running = 1;
}

/*
 * An adaptive channel hopper: the goal here is to maximize the time we 
 * spend on useful channels
 *
 * walk through the list of available channels
 * change to channels in even slots first, then odd (can randomize instead but no need I think)
 * for each channel, stay there for $timeslice milliseconds
 * if $timeslice is less than $min, skip channel
 * if $timeslice is more than $max, wait $max ms
 * after waiting, adjust $timeslice depending on received frames
 *  the more frames received, the more increase $timeslice gets
 * periodically normalize timeslices for each channel to avoid starvation
 * 
 */

int
run_hopper(hopper_t *h) {
  struct timeval t;
  pid_t parent;
  char line[DEF_LINELEN];
  fd_set fds;
  int i = 0;
  int delay_msec;
  int ch;
  mach_counter_t counter, counter2;
  int democracy = 0;

  while(1) {

    if(!h->channels[i].ch)
      i = 0;

    delay_msec = MIN_TIMESLICE + h->channels[i].bonus;
    if(delay_msec > MAX_TIMESLICE)
      delay_msec = MAX_TIMESLICE;
    
    mach_get_if_counters(h->device, &counter);

    t.tv_sec = delay_msec / 1000;
    t.tv_usec = delay_msec % 1000;
    FD_ZERO(&fds);
    FD_SET(h->sp[1], &fds);
    
    select(10, &fds, NULL, NULL, &t);

    if(FD_ISSET(h->sp[1], &fds)) {
      if(!fgets(line, DEF_LINELEN, h->rfp))
	continue;
      parse_command(h, line);
    }

    if(!h->running)
      continue;

    mach_get_if_counters(h->device, &counter2);
    if(counter2.rx-counter.rx > 0 && h->channels[i].bonus < MAX_TIMESLICE)
      h->channels[i].bonus += TIMESLICE_STEP;
    llog(3, "hopper: packet delta is %d, bonus for slot %d is %d\n", 
	 counter2.rx-counter.rx, i, h->channels[i].bonus);

    if(i < h->nchannels/2)
      ch = i*2;
    else if(i==h->nchannels-1)
      ch = i;
    else
      ch = 1 + (i - (h->nchannels/2))*2;

    if(h->channels[ch].ch) {
      llog(3, "hopper: setting channel to %d (index %d)\n", h->channels[ch].ch, ch);
      if(mach_channel_set(h->device, h->channels[ch].ch))
	llog(2, "hopper: mach_channel_set: %s\n", strerror(errno));
    }

    i++;

    parent = getppid();
    if(parent == 1 || kill(parent, 0)) {
      /* parent is gone. time for us to go */
      llog(2, "hopper: terminating\n");
      return 0;
    }

    if(democracy++ > RESET_PERIOD) {
      reset_bonus(h);
      democracy = 0;
    }
  }
  return 0;
}

hopper_t *
hopper_init(char *device) {
  hopper_t *h;
  int arg = 0;
  int i;
  
  if(!(h = xcalloc(sizeof(hopper_t), 1)))
    return NULL;

  if(socketpair(AF_UNIX, SOCK_STREAM, 0, h->sp)) {
    free(h);
    return NULL;
  }

  fcntl(h->sp[0], F_GETFL, arg);
  arg |= O_NONBLOCK;
  fcntl(h->sp[0], F_SETFL, arg);
  fcntl(h->sp[1], F_GETFL, arg);
  arg |= O_NONBLOCK;
  fcntl(h->sp[1], F_SETFL, arg);

  h->rfp = fdopen(h->sp[1], "r");
  h->wfp = fdopen(h->sp[0], "w");
  h->running = 0;
  strncpy(h->device, device, DEF_LINELEN);
  mach_get_channels(h->device, h->channels, MAX_CHANNELS);
  h->nchannels = 0;
  for(i=0; h->channels[i].ch; i++)
    h->nchannels++;
  llog(2, "hopper: %s supports %d channels total\n", h->device, h->nchannels);
  
  if(!(h->hopper_pid = fork())) {
    run_hopper(h);
    exit(0);
  }
  return h;
}

int
hopper_destroy(hopper_t *h) {
  int status;

  if(h == (hopper_t *)NULL)
    return 1;

  fclose(h->wfp);
  fclose(h->rfp);
  close(h->sp[0]);
  close(h->sp[1]);  
  kill(h->hopper_pid, SIGTERM);
  wait(&status);
  xfree(h);
  return 0;
}

int 
hopper_msg(hopper_t *h, char *msg) {
  if(h == (hopper_t *)NULL)
    return 1;

  fputs(msg, h->wfp);
  fflush(h->wfp);
  return 0;
}
