/*
 * ssidsniff - a tool for wireless network reconnaissance
 * Copyright (C) 2002-2007 Kostas Evangelinos <kos@bastard.net>
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*
 * $Id: mach-linux.c,v 1.17 2007-06-27 02:58:09 kos Exp $
 *
 * Bits taken from wireless-tools-27
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include "mach.h"
#include "util.h"

#ifdef HAVE_LINUX_WIRELESS_H
#include <linux/wireless.h>
#endif

typedef struct iw_range         iwrange;

#define KILO    1e3
#define MEGA    1e6
#define GIGA    1e9

typedef enum powersave_t {
  PS_NONE,
  PS_ACPI,
  PS_APM
} powersave_t;
static powersave_t powersave = PS_NONE;

static mach_channel_t ieee_80211_B_channels[] = {
  {1, 2412000000.0, 0},
  {2, 2417000000.0, 0},
  {3, 2422000000.0, 0},
  {4, 2427000000.0, 0},
  {5, 2432000000.0, 0},
  {6, 2437000000.0, 0},
  {7, 2442000000.0, 0},
  {8, 2447000000.0, 0},
  {9, 2452000000.0, 0},
  {10, 2457000000.0, 0},
  {11, 2462000000.0, 0},
  {0, 0.0, 0}
};

static double
iw_freq2float(const struct iw_freq *in) {
  int           i;
  double        res = (double) in->m;
  
  for(i = 0; i < in->e; i++)
    res *= 10;
  return(res);
}

int
iw_channel_to_freq(int                          channel,
                   double *                     pfreq,
                   const struct iw_range *      range)
{
  int           has_freq = 0;
  int           k;

  /* Check if the driver support only channels or if it has frequencies */
  for(k = 0; k < range->num_frequency; k++)
    {
      if((range->freq[k].e != 0) || (range->freq[k].m > (int) KILO))
        has_freq = 1;
    }
  if(!has_freq)
    return(-1);

  /* Find the correct frequency in the list */
  for(k = 0; k < range->num_frequency; k++)
    {
      if(range->freq[k].i == channel)
        {
          *pfreq = iw_freq2float(&(range->freq[k]));
          return(channel);
        }
    }
  /* Not found */
  return(-2);
}

int
iw_freq_to_channel(double                       freq,
                   const struct iw_range *      range)
{
  double        ref_freq;
  int           k;

  /* Check if it's a frequency or not already a channel */
  if(freq < KILO)
    return(-1);

  /* We compare the frequencies as double to ignore differences
   * in encoding. Slower, but safer... */
  for(k = 0; k < range->num_frequency; k++)
    {
      ref_freq = iw_freq2float(&(range->freq[k]));
      if(freq == ref_freq)
        return(range->freq[k].i);
    }
  /* Not found */
  return(-2);
}

double
mach_freq_get(char *device) {
  int skfd;
  struct iwreq wrq;
  double res;
  char buffer[sizeof(struct iw_range)*2];
  struct iw_range *range;
  int channel;
  
  skfd = socket(AF_INET, SOCK_DGRAM, 0);
  strcpy(wrq.ifr_name, device);
  if(ioctl(skfd, SIOCGIWFREQ, &wrq) != 0)
    res = -1;
  else 
    res = iw_freq2float(&wrq.u.freq);

  /* ewww */
  if(res < 1024) {
    channel = (int)res;
    wrq.u.data.pointer = (caddr_t) buffer;
    wrq.u.data.length = sizeof(buffer);
    wrq.u.data.flags = 0;
    strcpy(wrq.ifr_name, device);
    ioctl(skfd, SIOCGIWRANGE, &wrq);
    range = (struct iw_range *) buffer;
    
    iw_channel_to_freq(channel, &res, range);
  }

  close(skfd);
  return res;
}

static inline int
iw_get_ext(int                  skfd,           /* Socket to the kernel */
           const char *         ifname,         /* Device name */
           int                  request,        /* WE ID */
           struct iwreq *       pwrq)           /* Fixed part of the request */
{
  /* Set device name */
  strncpy(pwrq->ifr_name, ifname, IFNAMSIZ);
  /* Do the request */
  return(ioctl(skfd, request, pwrq));
}

/*
 * Get the range information out of the driver
 */
int
iw_get_range_info(int		skfd,
		  const char *	ifname,
		  iwrange *	range)
{
  struct iwreq		wrq;
  char			buffer[sizeof(iwrange) * 2];	/* Large enough */
  union iw_range_raw *	range_raw;

  /* Cleanup */
  bzero(buffer, sizeof(buffer));

  wrq.u.data.pointer = (caddr_t) buffer;
  wrq.u.data.length = sizeof(buffer);
  wrq.u.data.flags = 0;
  if(iw_get_ext(skfd, ifname, SIOCGIWRANGE, &wrq) < 0)
    return(-1);

  /* Point to the buffer */
  range_raw = (union iw_range_raw *) buffer;
  memcpy((char *) range, buffer, sizeof(iwrange));
  return(0);
}

int
mach_iswifi(char *device, char *descr, int len) {
  int skfd;
  struct iwreq wrq;
  int res;

  skfd = socket(AF_INET, SOCK_DGRAM, 0);
  strcpy(wrq.ifr_name, device);
  res = ioctl(skfd, SIOCGIWNAME, &wrq);
  if(descr)
    strncpy(descr, "Unknown", len); // XXX
  close(skfd);
  return res;
}

int
mach_channel_set(char *device, int channel) {
  int skfd;
  struct iwreq wrq;
  int res;
  
  memset((void *)&wrq, '\0', sizeof(struct iwreq));
  strcpy(wrq.ifr_name, device);
  wrq.u.freq.m = channel;
  skfd = socket(AF_INET, SOCK_DGRAM, 0);
  res = ioctl(skfd, SIOCSIWFREQ, &wrq);
  close(skfd);
  return res;  
}

char *
mach_battery_status(void) {
  static char line[MACH_LINELEN];
  FILE *fp;
  char *tokens[16];
  int ntokens;  
  struct stat st;
  int capacity, remaining;

  switch(powersave) {
  case PS_APM:
    if(!(fp = fopen("/proc/apm", "r")))
      return "";
    
    if(!fgets(line, MACH_LINELEN, fp)) {
      fclose(fp);
      return "";
    }
    fclose(fp);
    
    tokenize(line, tokens, &ntokens);
    if(ntokens < 8) {
      return "";
    }    
    sprintf(line, "Battery %s", tokens[6]);
    return line;
    break;
  case PS_ACPI:
    capacity = remaining = 0;
    if(!(fp = fopen("/proc/acpi/battery/BAT0/info", "r")))
      return "";
    
    while(fgets(line, MACH_LINELEN, fp)) {
      tokenize(line, tokens, &ntokens);
      if(!strcmp("last", tokens[0]) && !strcmp("full", tokens[1])) {
	sscanf(tokens[3], "%d", &capacity);
	break;
      }
    }
    fclose(fp);
    if(!capacity)
      return "";
    
    if(!(fp = fopen("/proc/acpi/battery/BAT0/state", "r")))
      return "";
    
    while(fgets(line, MACH_LINELEN, fp)) {
      tokenize(line, tokens, &ntokens);
      if(!strcmp("remaining", tokens[0])) {
	sscanf(tokens[2], "%d", &remaining);
	break;
      }
    }
    fclose(fp);
    if(!remaining)
      return "";
    
    sprintf(line, "Battery %d%%", remaining*100/capacity);
    return line;
    break;
  default:
    if(!stat("/proc/apm", &st)) {
      powersave = PS_APM;
      return mach_battery_status();
    }
    if(!stat("/proc/acpi", &st)) {
      powersave = PS_ACPI;
      return mach_battery_status();
    }
    return "";
    break;
  }
}

int
mach_set_if(char *device, int state) {
  int skfd;
  struct ifreq ifr;
  int res;

  skfd = socket(AF_INET, SOCK_DGRAM, 0); 
  strcpy(ifr.ifr_name, device);
  res = ioctl(skfd, SIOCGIFFLAGS, &ifr);
  ifr.ifr_flags |= IFF_UP;
  res += ioctl(skfd, SIOCSIFFLAGS, &ifr);
  close(skfd);
  return res;
}

int
mach_set_monitor(char *device, int state) {
  int skfd;
  struct iwreq wrq;
  int res;

  skfd = socket(AF_INET, SOCK_DGRAM, 0); 
  strcpy(wrq.ifr_name, device);
  res = ioctl(skfd, SIOCGIWMODE, &wrq);
  if(state)
    wrq.u.mode = IW_MODE_MONITOR;
  else
    wrq.u.mode = IW_MODE_INFRA;
  res += ioctl(skfd, SIOCSIWMODE, &wrq);
  close(skfd);
  return res;
}

#undef  MIN
#define MIN(x,y) ((x) < (y) ? (x) : (y))

#undef  MAX
#define MAX(x,y) ((x) > (y) ? (x) : (y))

/* iwlist.c:print_freq_info */
/* when in doubt, use 80211B rather than failing */
int
mach_get_channels(char *ifname, mach_channel_t *channels, int maxl) {
  struct iw_range       range;
  double                freq;
  int                   k;
  int skfd;

  memset(channels, '\0', maxl*sizeof(mach_channel_t));
  skfd = socket(AF_INET, SOCK_DGRAM, 0); 
  
  if(iw_get_range_info(skfd, ifname, &range) < 0) {
    memcpy(channels, &ieee_80211_B_channels, MIN(maxl, 12)*sizeof(mach_channel_t));
    close(skfd);
    return 0;
  }

  close(skfd);
  if(range.num_frequency <= 0) {
    memcpy(channels, &ieee_80211_B_channels, MIN(maxl, 12)*sizeof(mach_channel_t));
    return 0;
  }

  for(k = 0; k < range.num_frequency; k++) {    
    if(k>maxl)
      break;
    freq = iw_freq2float(&(range.freq[k]));
    channels[k].ch = range.freq[k].i;
    channels[k].freq = freq;
    channels[k].bonus = 0;
  }
  return(0);
}

int
mach_get_if_counters(char *ifname, mach_counter_t *c) {
  static char line[MACH_LINELEN];
  FILE *fp;
  char *tokens[32];
  int ntokens;  

  if(!(fp = fopen("/proc/net/dev", "r")))
    return 1;
  
  while(fgets(line, MACH_LINELEN, fp)) {
    if(strstr(line, ifname)) {
      tokenize(line, tokens, &ntokens);
      c->rx = atoi(tokens[2]);
      c->tx = atoi(tokens[10]);
      fclose(fp);
      return 0;
    }
  }

  fclose(fp);
  return 3;
}
