/*
 * ssidsniff - a tool for wireless network reconnaissance
 * Copyright (C) 2002-2007 Kostas Evangelinos <kos@bastard.net>
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*
 * $Id: pcapdump.c,v 1.14 2007-06-23 16:47:43 kos Exp $
 *
 * A testbed for the 80211 dissector
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#if HAVE_GETOPT_H
#include <getopt.h>
#endif

#include "llog.h"
#include "80211.h"
#include "mach.h"
#include "support.h"

#define DEF_LINELEN 1024
int running = 1;
int interrupted = 0;
int live;
extern char pcap_version[];

void usage(char *argv0, int code) {
  fprintf(stderr, "Usage: %s [-i interface] [-r filename] [-v level] [-f filter]\n", argv0);
  fprintf(stderr, "\t-i interface  Set the interface to use\n");
  fprintf(stderr, "\t-r filename   Set capture filename (PCAP format)\n");
  fprintf(stderr, "\t-v level      Set verbose level\n");
  fprintf(stderr, "\t-f filter     Set PCAP filter expression\n");
  exit(code);
}

void
signal_handler(int sig) {
  switch(sig){
  case SIGINT:
    interrupted = 1;
  }
}

int
do_pcap_read(pcap_t *pd, int snaplen, char *filter) {
  struct pcap_pkthdr pkthdr;
  char *pkt;
  unsigned int count;
  f80211_t frame;
  int dlt;
  
  dlt = pcap_datalink(pd);
  count = 0;

  while(running) {
    if(interrupted)
      break;
  
    if(!(pkt = (char *)pcap_next(pd, &pkthdr))) {
      llog(0, "ssid_sniff: pcap_next: %s\n", pcap_geterr(pd));
      break;
    }
    
    if(pkthdr.caplen != pkthdr.len) {
      llog(2, "pcapdump: warning: truncated packet\n");
    }
    
    llog(1, "> frame %d, length %u (0x%x), timestamp %u:%u\n",
	 count, pkthdr.len, pkthdr.len, pkthdr.ts.tv_sec, pkthdr.ts.tv_usec);    
    llog_hexdump(9, pkt, pkthdr.len);
    f80211_parse((unsigned char *)pkt, &pkthdr, dlt, &frame);
    count++;
  }
  return 0;
}

int
main(int argc, char **argv) {
  char c;
  int snaplen;
  char fname[DEF_LINELEN];
  int verbose;
  char *pdev;
  char filter[DEF_LINELEN];
  char device[DEF_LINELEN];
  pcap_t *pd;
  struct bpf_program pcap_fcode;
  unsigned int pcap_if_net;
  unsigned int pcap_if_mask;
  char pcap_ebuf[PCAP_ERRBUF_SIZE];
  
  running = 1;
  live = 1;
  snaplen = 2500;
  verbose = 1;
  fname[0] = '\0';
  filter[0] = '\0';
  device[0] = '\0';

  while((c = getopt(argc, argv, "r:v:f:i:")) != EOF) {
    switch(c) {
    case 'i':
      strncpy(device, optarg, DEF_LINELEN);
      live = 1;
      break;
    case 'r':
      strncpy(fname, optarg, DEF_LINELEN);
      live = 0;
      break;
    case 'v':
      verbose = atoi(optarg);
      break;
    case 'f':
      strncpy(filter, optarg, DEF_LINELEN);
      break;
    case '?':
      usage(argv[0], 1);
      break;
    }
  }

  llog_init(LLOG_STDERR);
  llog_level(verbose);
  signal(SIGINT, signal_handler);
  llog(1, "%s: libpcap version %s\n", argv[0], pcap_version);

  if(live) {
    if(!*device) {
      if(!(pdev = pcap_lookupdev(device))) {
	llog(0, "%s: No suitable device found\n", argv[0]);
	exit(1);
      }
      strncpy(device, pdev, PCAP_ERRBUF_SIZE);
      llog(1, "%s: Detected device %s\n", argv[0], device);
    } else {
      llog(1, "%s: Using device %s\n", argv[0], device);
    }

    mach_set_if(device, 1);
    mach_set_monitor(device, 1);
    
    if(!(pd = pcap_open_live(device, snaplen, 1, 100, pcap_ebuf))) {
      llog(0, "pcap_open_live(%s): %s\n", device, pcap_ebuf);
      exit(2);
    }
    pcap_lookupnet(device, &pcap_if_net, &pcap_if_mask, pcap_ebuf);
    pcap_compile(pd, &pcap_fcode, filter, 1000, pcap_if_mask);
    pcap_setfilter(pd, &pcap_fcode);       

    if(mach_iswifi(device, NULL, 0)) {
      llog(1, "%s: Device %s isn't a wireless interface, continuing anyway\n", argv[0], device);
      sleep(2);
    }

    if(pcap_set_datalink(pd, DLT_IEEE802_11_RADIO) == -1)
      if(pcap_set_datalink(pd, DLT_PRISM_HEADER) == -1)
	pcap_set_datalink(pd, DLT_IEEE802_11);
    
    if(f80211_checkdlt(pcap_datalink(pd))) {
      llog(1, "%s: Link type %s isn't compatible, continuing anyway\n",
	   argv[0],
	   pcap_datalink_val_to_name(pcap_datalink(pd)));
      sleep(2);
    }
  } else {
    if(!(pd = pcap_open_offline(fname, pcap_ebuf))) {
      llog(0, "pcap_open_offline(%s): %s\n", fname, pcap_ebuf);
      exit(4);
    }      
  }

  do_pcap_read(pd, snaplen, filter);

  pcap_freecode(&pcap_fcode);
  pcap_close(pd);

  llog_close();
  exit(0);
}

