/*
 * ssidsniff - a tool for wireless network reconnaissance
 * Copyright (C) 2002-2007 Kostas Evangelinos <kos@bastard.net>
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*
 * $Id: ssidsniff.c,v 1.78 2007-06-28 14:28:01 kos Exp $
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>

#if HAVE_GETOPT_H
#include <getopt.h>
#endif

#include "xmalloc.h"
#include "llog.h"
#include "ouidb.h"
#include "xpaudio.h"
#include "ssidsniff.h"
#include "util.h"
#include "80211.h"
#include "mach.h"
#include "cursesui.h"
#include "support.h"

#ifdef HAVE_PCAP_H
#include "pcap.h"
#endif

#ifdef HAVE_NET_BPF_H
#include "net/bpf.h"
#endif

ssidsniff_t *gs;

RETSIGTYPE
signal_handler(int sig) {
  switch(sig){
  case SIGINT:
    gs->running = 0;
    break;
  case SIGALRM:
    alarm(gs->delay);
    pps_update_all(gs);
    /* fall through */
  case SIGWINCH:
  case SIGCONT:
    ui_redraw(gs);    
    break;
  }
  signal(sig, signal_handler);
}

int
counter_clear(counter_t *count) {
  count->total = count->beacon = count->data = count->martian = 0;
  return 0;
}

/* make noise */
int
geiger_packet(ssidsniff_t *s, struct pcap_pkthdr *h, char *sp) {
  
  if(s->geiger && !s->xpa) {
    if(!(s->xpa = xpaudio_new(NULL))) {
      ui_info("Failed to start geiger mode");
      s->geiger = 0;
      return 1;
    }
    ui_info("Geiger mode on: %s", xpaudio_describe(s->xpa));
  }

  if(s->xpa)
    xpaudio_send(s->xpa, sp, h->len>DEF_GEIGER_L?DEF_GEIGER_L:h->len);    
  return 0;
}

int
default_writefile(char *p, int len, char *type) {
  struct tm *tm;
  char hostname[DEF_LINELEN];
  int serial = 0;
  struct stat st;
  time_t now;
  
  if(gethostname(hostname, DEF_LINELEN)) {
    strncpy(p, DEF_WRFILE, len);
    return 2;
  }

  now = time(NULL);
  if(!(tm = localtime(&now))) {
    strncpy(p, DEF_WRFILE, len);
    return 1;
  }
  
  while(serial++ < 1000) {
    if(type)
      snprintf(p, len, "%s-%s-%04d-%02d-%02d-%02d", hostname, type, tm->tm_year+1900, 
	       tm->tm_mon+1, tm->tm_mday, serial);
    else
      snprintf(p, len, "%s-%04d-%02d-%02d-%02d.cap", hostname, tm->tm_year+1900, 
	       tm->tm_mon+1, tm->tm_mday, serial);
    
    if(stat(p, &st))
      break;
  }
  
  return 0;
}

/* lazy version of pcap_dump */
int
save_packet(ssidsniff_t *s, struct pcap_pkthdr *h, char *sp) {
  
  if(s->dumpfile[0] && !s->pd_dump) {
    default_writefile(s->dumpfile, DEF_LINELEN, NULL);
    if(!(s->pd_dump = pcap_dump_open(s->pd, s->dumpfile))) {
      ui_info("pcap_dump_open(%s): %s", s->dumpfile, pcap_geterr(s->pd));
      s->dumpfile[0] = '\0';
      return 1;
    }
    ui_info("Capture file is %s", s->dumpfile);
  }

  pcap_dump((unsigned char *)s->pd_dump, h, (unsigned char *)sp);
  return 0;
}

double
timeval_diff(struct timeval t1, struct timeval t2) {
  double diff;
  
  diff = ((double)t2.tv_usec - (double)t1.tv_usec)/1000000 + 
    ((double)t2.tv_sec - (double)t1.tv_sec);
  return fabs(diff);
}

int
pps_update_all(ssidsniff_t *s) {
  struct timeval now;
  struct timezone tz;
  int i;
  wnet_t *n;
  double pps_cur;

  gettimeofday(&now, &tz);
  for(i=0; i<s->num_networks; i++) {
    n = &s->networks[i];
    pps_cur = (double)n->pps_delta / timeval_diff(n->pps_prevts, now);
    n->pps = (int)(((double)n->pps + pps_cur) / 2);
    n->pps_prevts = now;
    n->pps_delta = 0;
  }
  return 0;
}

int
pps_update_net(wnet_t *n) {

  n->pps_delta++;
  return 0;
}

int
wnet_clear(wnet_t *res) {
  memset((void *)res, '\0', sizeof(wnet_t));
  res->signal = CHAR_MIN;
  counter_clear(&res->count);
  strings_clear(&res->strings);
  return 0;
}

/* Only create a unique entry, don't enrich */
wnet_t *
wnet_add(ssidsniff_t *s, char *key) {
  wnet_t *n;
  void *p;

  if(s->num_networks == s->max_networks) {
    if(!(p = xrealloc(s->networks, (s->max_networks + DEF_WNET_ALLOC) * sizeof(wnet_t))))
      return (wnet_t *)NULL;
    s->max_networks += DEF_WNET_ALLOC;
    s->networks = (wnet_t *)p;
  }

  n = &s->networks[s->num_networks];
  wnet_clear(n);
  strncpy(n->bss, key, DEF_LINELEN);
  n->serial = s->num_networks;
  s->num_networks++;
  return n;
}

int
wnet_client_add(wnet_t *n, char *key) {
  wnet_client_t *c, *p;

  if(!(c = xcalloc(1, sizeof(wnet_client_t))))
    return 1;
  strncpy(c->src, key, DEF_LINELEN);
  c->next = (wnet_client_t *)NULL;
  if(!n->clients) {
    n->clients = c;
    return 0;
  }
  
  for(p=n->clients; p->next; p=p->next)
    ;
  p->next = c;
  return 0;
}

int
wnet_nclients(wnet_t *n) {
  int i;
  wnet_client_t *c;

  if(!n->clients)
    return 0;
  for(c=n->clients,i=0; c->next; c=c->next)
    i++;
  return i;
}

wnet_client_t *
wnet_client_find(wnet_t *n, char *key) {
  wnet_client_t *c;

  if(!n->clients)
    return (wnet_client_t *)NULL;
  for(c=n->clients; c; c=c->next)
    if(!strcmp(c->src, key))
      return c;
  return (wnet_client_t *)NULL;
}

int
wnet_client_logpacket(wnet_t *n, f80211_t *frame) {
  wnet_client_t *c;
  
  if(!(c = wnet_client_find(n, frame->src)))
    wnet_client_add(n, frame->src);
  assert((c = wnet_client_find(n, frame->src)));

  if(c->manufacturer[0] == '\0')
    ouidb_get_vendor(frame->src, c->manufacturer, DEF_LINELEN);
  if(!c->ap && frame->ap)
    c->ap = 1;
  c->count++;
  return 0;
}

int
wnet_record(ssidsniff_t *s) {
  int i;
  char fname[DEF_LINELEN];
  FILE *fp;
  wnet_t *n;

  default_writefile(fname, DEF_LINELEN, "NetworkList");
  if(!(fp = fopen(fname, "w"))) {
    ui_info("Failed to record network list: %s", strerror(errno));
    return 1;
  }
  fprintf(fp, "BSSID,SSID,Channel,WEPCapable,WEPUsed,IsAP,SignalStrength,CountPackets,CountData\n");

  for(i=0; i<s->num_networks; i++) {
    n = &s->networks[i];
    fprintf(fp, "%s,%s,%d,%d,%d,%d,%d,%d,%d\n",
	    n->bss,
	    n->ssid,
	    n->channel,
	    n->wepcapable,
	    n->wepused,
	    n->ap,
	    n->signal,
	    n->count.total,
	    n->count.data
	    );
  }
  fclose(fp);
  ui_info("Recorded networks to %s", fname);
  return 0;  
}

/* Enrich network struct with whatever we find */
int
wnet_logpacket(ssidsniff_t *s, wnet_t *network, f80211_t *frame, 
	       struct pcap_pkthdr *pkthdr, char *pkt) {

  pps_update_net(network);
  wnet_client_logpacket(network, frame);
  network->last_activity = pkthdr->ts;

  switch(f80211_type(frame)) {
  case f80211_BEACON:
    s->globalcount.beacon++;
    network->count.beacon++;
    break;
  case f80211_DATA:
    s->globalcount.data++;
    network->count.data++;
    strings_add(&network->strings, pkt, pkthdr->len);
    if(s->capture == CAPTURE_DATA || network->capture)
      save_packet(s, pkthdr, pkt);
    break;
  case f80211_PROBERESP:
  case f80211_PROBEREQ:
  default:
    s->globalcount.martian++;
    network->count.martian++;
    break;
  }

  if(!network->freq)
    network->freq = mach_freq_get(s->device);
  if(!network->channel)
    network->channel = frame->channel;
  if(network->dst[0] == '\0')
    strncpy(network->dst, frame->dst, DEF_LINELEN);
  if(strlen(frame->ssid))
    strncpy(network->ssid, frame->ssid, DEF_LINELEN);
  if(!network->ap && frame->ap)
    network->ap = 1;
  if(!network->wepcapable && frame->wepcapable)
    network->wepcapable = 1;
  if(!network->wepused && frame->wepused)
    network->wepused = 1;
  if(!strcmp(network->bss, frame->src) && frame->signal)
    network->signal = frame->signal;
  if(!network->xrates && frame->xrates)
    network->xrates = 1;
  if(!network->wpa && frame->wpa)
    network->wpa = 1;
  if(!network->wpa2 && frame->wpa2)
    network->wpa2 = 1;
  
  return 0;
}

wnet_t *
wnet_find(ssidsniff_t *s, char *which) {
  int i;
  wnet_t *n;

  for(i=0; i<s->num_networks; i++) {
    n = &s->networks[i];
    if(!strcmp(n->bss, which))
      return n;
  }
  return (wnet_t *)NULL;
}

int
process_packet(ssidsniff_t *s, struct pcap_pkthdr pkthdr, char *pkt) {
  f80211_t frame;
  wnet_t *network;
  
  s->globalcount.total++;

  if(s->geiger)
    geiger_packet(s, &pkthdr, pkt);

  if(s->capture == CAPTURE_ALL) 
    save_packet(s, &pkthdr, pkt);
  
  if(f80211_parse((unsigned char *)pkt, &pkthdr, s->pcap_dlt, &frame)) {
    s->globalcount.martian++;
    return 3;
  }

  if(!strlen(frame.bss) || !strcmp(frame.bss, "ff:ff:ff:ff:ff:ff"))
    return 4;

  /* is this new? */
  network = wnet_find(s, frame.bss);
  if(!network) {
    network = wnet_add(s, frame.bss);
    ui_info("New network %s", frame.bss);
    llog(2, "New network %s\n", frame.bss);
  } 

  assert(network != (wnet_t *)NULL);
  network->count.total++;

  wnet_logpacket(s, network, &frame, &pkthdr, pkt);  
  return 0;
}

int
ssid_sniff_live(ssidsniff_t *s, char *argv0) {
  fd_set fds;
  int pcap_promiscuous = 1;
  struct bpf_program pcap_fcode;
  unsigned int pcap_if_net;
  unsigned int pcap_if_mask;
  char pcap_ebuf[PCAP_ERRBUF_SIZE];
  char *pdev = NULL;
  struct pcap_pkthdr pkthdr;
  char *pkt;
  int prevdlt;

  /* Find network device */
  if(!*s->device) {
    if(!(pdev = pcap_lookupdev(s->device))) {
      llog(0, "%s: No suitable device found\n", argv0);
      exit(1);
    }
    strncpy(s->device, pdev, PCAP_ERRBUF_SIZE);
  }
  llog(1, "%s: Using device %s\n", argv0, s->device);

  mach_set_if(s->device, 1);
  mach_set_monitor(s->device, 1);

  /* Initialize hopper */
  s->hopper = hopper_init(s->device);
  if(s->run_hopper)
    hopper_msg(s->hopper, "start");

  /* Initialize capturing */
  if(!(s->pd = pcap_open_live(s->device, 2500, pcap_promiscuous, 100, pcap_ebuf))) {
    llog(0, "pcap_open_live(%s): %s\n", s->device, pcap_ebuf);
    return 2;
  }
  s->pcap_fd = pcap_fileno(s->pd);
  prevdlt = pcap_datalink(s->pd);

  pcap_lookupnet(s->device, &pcap_if_net, &pcap_if_mask, pcap_ebuf);
  if(pcap_compile(s->pd, &pcap_fcode, s->filter, 1000, pcap_if_mask) == -1) {
    llog(1, "%s: filter %s: pcap_compile failed.\n", s->device, s->filter);
    return 10;
  }

  if(pcap_setfilter(s->pd, &pcap_fcode) == -1) {
    llog(1, "%s: filter %s: pcap_setfilter failed.\n", s->device, s->filter);
    return 11;
  }

  if(mach_iswifi(s->device, NULL, 0)) {
    llog(1, "%s: Device %s isn't a wireless interface, continuing anyway\n", argv0, s->device);
    sleep(2);
  }

  if(pcap_set_datalink(s->pd, DLT_IEEE802_11_RADIO) == -1)
    if(pcap_set_datalink(s->pd, DLT_PRISM_HEADER) == -1)
      pcap_set_datalink(s->pd, DLT_IEEE802_11);

  if(f80211_checkdlt(pcap_datalink(s->pd))) {
    llog(1, "%s: Link type %s isn't compatible, continuing anyway\n",
	 argv0,
	 pcap_datalink_val_to_name(pcap_datalink(s->pd)));
    sleep(2);
  }
  s->pcap_dlt = pcap_datalink(s->pd);
  
  llog(1, "%s: Starting capture with filter=%s on %s\n", argv0, s->filter, s->device);
  ui_init();
  counter_clear(&s->globalcount);
  signal(SIGINT, signal_handler);
  signal(SIGCONT, signal_handler);
  signal(SIGALRM, signal_handler);  
  alarm(s->delay);
  ui_redraw(s);

  while(s->running) {

    FD_ZERO(&fds);
    FD_SET(s->pcap_fd, &fds);
    FD_SET(0, &fds);
    
    if(select(s->pcap_fd+1, &fds, NULL, NULL, NULL) < 0) {
      if(errno == EINTR)
	continue;
      else
	break;
    }

    if(FD_ISSET(0, &fds)) {
      ui_process_events(s);
      continue;
    }

    if(!FD_ISSET(s->pcap_fd, &fds))
      continue;
  
    if(!(pkt = (char *)pcap_next(s->pd, &pkthdr))) {
      ui_info("pcap_next: %s\n", pcap_geterr(s->pd));      
      continue;
    }
    
    process_packet(s, pkthdr, pkt);
  }
  
  alarm(0);

  /* stop hopper */
  if(s->hopper) {
    hopper_destroy(s->hopper);
    s->hopper = (hopper_t *)NULL;
  }
  
  if(s->pd_dump)
    pcap_dump_close(s->pd_dump);

  pcap_set_datalink(s->pd, prevdlt);
  pcap_freecode(&pcap_fcode);
  pcap_close(s->pd);
  
  /* restore interface */
  mach_set_monitor(s->device, 0);
  mach_set_if(s->device, 0);

  ui_finish();
  return 0;
}

int
ssid_sniff_file(ssidsniff_t *s, char *argv0, char *pcap_file) {
  char pcap_ebuf[PCAP_ERRBUF_SIZE];
  struct pcap_pkthdr pkthdr;
  char *pkt;

  sprintf(s->device, "<file>");
    
  if(!(s->pd = pcap_open_offline(pcap_file, pcap_ebuf))) {
    llog(0, "%s: pcap_open_offline: %s\n", argv0, pcap_ebuf);
    return 2;
  }

  ui_init();
  counter_clear(&s->globalcount);
  signal(SIGINT, signal_handler);
  signal(SIGCONT, signal_handler);
  signal(SIGALRM, signal_handler);  
  alarm(s->delay);
  ui_redraw(s);
  s->pcap_dlt = pcap_datalink(s->pd);

  while((pkt = (char *)pcap_next(s->pd, &pkthdr)))
    process_packet(s, pkthdr, pkt);

  pcap_close(s->pd);
  while(s->running) {
    ui_process_events(s);
  }

  alarm(0);
  ui_finish();
  return 0;
}

void
usage(char *argv0, int code) {
  /*               ------------------------------------------------------------------------------ */
  fprintf(stderr, "Usage: %s [-VdDg] [-i device] [-f filter] [-v level]\n", argv0);
  fprintf(stderr, "       [-r file] [-w file]\n\n");
  fprintf(stderr, "\t-V              display version number and exit\n");
  fprintf(stderr, "\t-d              Save all data packets received\n");
  fprintf(stderr, "\t-D              Save all packets received\n");
  fprintf(stderr, "\t-i interface    Set the interface to listen on\n");
  fprintf(stderr, "\t-f filter       PCAP filter expression to apply\n");
  fprintf(stderr, "\t-g              Geiger counter mode; enables sound for received packets\n");
  fprintf(stderr, "\t-r filename     Read from PCAP file instead of using a network device\n");
  fprintf(stderr, "\t-w filename     Set PCAP capture file name\n");
  fprintf(stderr, "\t-H              Hop through channels automatically\n");
  fprintf(stderr, "\t-v verbosity    set debugging level; messages go to standard error\n");
  exit(code);
}

int
main(int argc, char **argv) {
  char c;
  int verbose;
  char file[DEF_LINELEN];
  ssidsniff_t s;
  
  /* defaults */
  verbose = 1;
  file[0] = '\0';

  s.device[0] = '\0';
  s.geiger = 0;
  s.dumpfile[0] = '\0';
  s.filter[0] = '\0';
  s.xpa = (xpaudio_t *)NULL;
  s.num_networks = 0;
  s.max_networks = DEF_WNET_ALLOC;
  s.start = time(NULL);
  s.delay = 1;
  s.running = 1;
  default_writefile(s.dumpfile, DEF_LINELEN, NULL);
  s.pd_dump = (pcap_dumper_t *)NULL;
  s.capture = CAPTURE_NONE;
  s.hopper = (hopper_t *)NULL;
  s.run_hopper = 0;
  gs = &s;

  if(!(s.networks = (wnet_t *)xcalloc(s.max_networks, sizeof(wnet_t)))) {
    fprintf(stderr, "%s: Error allocating memory\n", argv[0]);
    exit(1);
  }

  while((c = getopt(argc, argv, "i:f:gw:v:VdDHr:"))!=EOF) {
    switch(c) {
    case 'i':
      strncpy(s.device, optarg, PCAP_ERRBUF_SIZE);
      break;
    case 'v':
      verbose = atoi(optarg);
      break;
    case 'f':
      strncpy(s.filter, optarg, DEF_LINELEN);
      break;
    case 'g':
      s.geiger = 1;
      break;
    case 'w':
      strncpy(s.dumpfile, optarg, DEF_LINELEN);
      break;
    case 'V':
      printf("%s %s\n", PACKAGE, VERSION);
      exit(0);
    case 'd':
      s.capture = CAPTURE_DATA;
      break;
    case 'D':
      s.capture = CAPTURE_ALL;
      break;
    case 'H':
      s.run_hopper = 1;
      break;
    case 'r':
      strncpy(file, optarg, DEF_LINELEN);
      break;
    case '?':
      usage(argv[0], 1);
      break;
    }
  }

  llog_init(LLOG_STDERR);
  llog_level(verbose);

  if(file[0] == '\0')
    ssid_sniff_live(&s, argv[0]);  
  else
    ssid_sniff_file(&s, argv[0], file);
 
  if(s.xpa)
    xpaudio_destroy(s.xpa);
    
  llog_close();
  exit(0);
}

