/*
 * ssidsniff - a tool for wireless network reconnaissance
 * Copyright (C) 2002-2007 Kostas Evangelinos <kos@bastard.net>
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*
 * $Id: xpaudio-alsa.c,v 1.3 2007-06-20 01:47:12 kos Exp $
 *
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/param.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#ifdef HAVE_ALSA_ASOUNDLIB_H
#include <alsa/asoundlib.h>
#endif

#include "xmalloc.h"
#include "xpaudio.h"
#include "llog.h"
#include "cursesui.h"

#define XPA_DEFAULTDEVICE "plughw:0,0"

xpaudio_t *
xpaudio_new(char *ud) {
  xpaudio_t *xpa;
  int err;
	
  xpa = xcalloc(1, sizeof(xpaudio_t));
  if(ud)
    strncpy(xpa->device, ud, XPA_LINELEN);
  else
    strncpy(xpa->device, XPA_DEFAULTDEVICE, XPA_LINELEN);
  xpa->rate = XPA_RATE_DEFAULT;
  xpa->channels = XPA_NCHANNELS_DEFAULT;

  if((err = snd_pcm_open(&xpa->alsa_handle, xpa->device, SND_PCM_STREAM_PLAYBACK, 0)) < 0) {
    ui_info("Cannot open audio device %s: %s\n", xpa->device, snd_strerror(err));
    free(xpa);
    return (xpaudio_t *)NULL;
  }
  
  if((err = snd_pcm_hw_params_malloc(&xpa->hw_params)) < 0) {
    llog(3, "cannot allocate hardware parameter structure (%s)\n", snd_strerror (err));
    free(xpa);
    return (xpaudio_t *)NULL;
  }
				 
  if((err = snd_pcm_hw_params_any(xpa->alsa_handle, xpa->hw_params)) < 0)
    llog(3, "cannot initialize hardware parameter structure (%s)\n", snd_strerror(err));
  
  if((err = snd_pcm_hw_params_set_access(xpa->alsa_handle, xpa->hw_params, SND_PCM_ACCESS_RW_INTERLEAVED)) < 0)
    llog(3, "cannot set access type (%s)\n", snd_strerror(err));
	
  if((err = snd_pcm_hw_params_set_format(xpa->alsa_handle, xpa->hw_params, SND_PCM_FORMAT_S16_LE)) < 0)
    llog(3, "cannot set sample format (%s)\n", snd_strerror(err));
	
  if((err = snd_pcm_hw_params_set_rate_near(xpa->alsa_handle, xpa->hw_params, &xpa->rate, 0)) < 0)
    llog(3, "cannot set sample rate (%s)\n", snd_strerror(err));
  
  if((err = snd_pcm_hw_params_set_channels(xpa->alsa_handle, xpa->hw_params, xpa->channels)) < 0)
    llog(3, "alsa: cannot set channel count (%s)\n", snd_strerror(err));
  
  if((err = snd_pcm_hw_params(xpa->alsa_handle, xpa->hw_params)) < 0)
    llog(3, "alsa: cannot set parameters (%s)\n", snd_strerror(err));
  
  if((err = snd_pcm_prepare(xpa->alsa_handle)) < 0) {
    llog(3, "cannot prepare audio interface for use (%s)\n", snd_strerror (err));
    free(xpa);
    return (xpaudio_t *)NULL;
  }
  
  return xpa;
}

int
xpaudio_destroy(xpaudio_t *xpa) {
  snd_pcm_drop(xpa->alsa_handle);
  snd_pcm_close(xpa->alsa_handle);  
  snd_pcm_hw_params_free (xpa->hw_params);
  xfree(xpa);
  return 0;
}

int
xpaudio_set(xpaudio_t *xpa, unsigned int param, int val) {
  int ret;
  int err;

  switch(param) {
  case XPA_RATE:    
    xpa->rate = val;
    if((err = snd_pcm_hw_params_set_rate_near(xpa->alsa_handle, xpa->hw_params, &xpa->rate, 0)) < 0)
      llog(3, "cannot set sample rate (%s)\n", snd_strerror(err));
    break;
  case XPA_NCHANNELS:
    xpa->channels = val;
    if((err = snd_pcm_hw_params_set_channels(xpa->alsa_handle, xpa->hw_params, xpa->channels)) < 0)
      llog(3, "alsa: cannot set channel count (%s)\n", snd_strerror(err));
    break;
  case XPA_ENCODING:
    xpa->encoding = val;
    break;
  default:
    return -1;
  }
  
  return ret;
}

int 
xpaudio_get(xpaudio_t *xpa, unsigned int param) {
  
  switch(param) {
  case XPA_RATE:
    return(xpa->rate);
    break;
  case XPA_NCHANNELS:
    return(xpa->channels);
    break;
  case XPA_ENCODING:
    return(xpa->encoding);
    break;
  default:    
    return 1;
  }  
  return 0;
}

char *
xpaudio_describe(xpaudio_t *xpa) {
  return "ALSA sound module";
}

int
xpaudio_send(xpaudio_t *xpa, char *data, unsigned int len) {
  int i;
  char buffer[XPA_FRAGSIZE];

  memset(buffer, '\0', XPA_FRAGSIZE);
  memcpy(buffer, data, MIN(len, XPA_FRAGSIZE));
  
  while((i = snd_pcm_writei(xpa->alsa_handle, buffer, XPA_FRAGSIZE)) < 0) {
    snd_pcm_prepare(xpa->alsa_handle);
  }

  return 0;
}
